
%if 0

RENUMBER - Change resident AMIS multiplexer's multiplex number
 2021 by C. Masloch

Usage of the works is permitted provided that this
instrument is retained with the works, so that any entity
that uses the works is notified of this instrument.

DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.

%endif

	%assign __lMACROS1_MAC__DEBUG_DEFAULTS	1
%include "lmacros3.mac"
%include "AMIS.MAC"

	defaulting

		; Option flags (hold in bp)
_force		equ 1
_got1stnumber	equ 100h
_got2ndnumber	equ 200h

%ifdef _MAP
	[map all _MAP]
%endif

	cpu 8086
	org 100h


transient:
			; This is a simple MS-DOS 1.x detection (8 Byte).
	mov ah, 4Dh		; the only function I found that doesn't need much setup or cleanup
	stc			; but that uses MS-DOS 2.x error reporting. clears CF (Int21.30 does not)
	int 21h
	jnc .dosgood		; it is MS-DOS 2.x+ or compatible -->
	retn			; it isn't so simply terminate
.dosgood:
	cld
	mov dx, msg.name
	call disp_msg		; Start with message

			; The UMB link and memory allocation strategy is now always safed here
			; and restored by abort_msg as well as any other termination code. This
			; saves any code changing the strategy the effort to restore it proper.
	mov ax, 5802h
	int 21h
	xor ah, ah		; some "DOS" only return al
	mov word [ restored.umblink ], ax	; save UMB link

	mov ax, 5800h
	int 21h
	mov word [ restored.strategy ], ax	; save strategy

	mov dx, i23
	mov ax, 2523h
	int 21h			; set Int23 so that we restore link/strategy on Ctrl-C as well
	mov dx, i24
	mov ax, 2524h
	int 21h			; set Int24 so all hard errors are changed to the soft error

getfirstmcb:
	mov ah, 52h
	int 21h
	mov ax, word [ es:bx - 2 ]
	mov word [firstmcb], ax

getfirstumcb:
			; We try to get the first UMCB for various uses later. Note that we
			; never use the result as an MCB address so if it's wrong we won't
			; crash or something. (We compare memory block or MCB addresses to
			; the value to check whether they're in the UMA.)
	mov ax, 1261h		; PTS-DOS: Get first UMCB
	stc
	int 2Fh
	jc .determine		; not supported -->
	inc ax
	cmp ax, byte 2		; -1, 0, 1 ?
	jbe .determine		; not supported (or none) -->
	dec ax
	mov word [ firstumcb ], ax	; set UMB
	jmp short .got		; got it -->

.determine:
	mov ax, 5803h
	xor bx, bx
	int 21h			; disable UMB link, leave only LMA chain
	jc .none		; that isn't supported either -->

	mov ax, word [firstmcb]
	mov dx, ax		; first MCB
	xor bx, bx		; use offsets from bx, not addresses
.looplmb:
	mov ds, ax
	inc ax
	add ax, word [ bx + 3 ]	; next MCB's address
	cmp byte [ bx ], 'M'
	je .looplmb		; not last -->
	cmp byte [ bx ], 'Z'
	jne .none		; corrupted -->
	xchg ax, dx		; dx = what we assume to be the first UMA chain MCB
				; ax = first MCB

	push ax
	inc bx			; = 1
	mov ax, 5803h
	int 21h			; enable UMB link, include UMA chain
	pop ax
	jc .none		; so we can disable it but not enable? -->

	dec bx			; = 0
	xor cx, cx		; flag if assumed first UMCB found
.loopumb:
	cmp ax, dx
	jne .notlastlmb
	inc cx			; there it is
.notlastlmb:
	mov ds, ax
	cmp byte [ bx ], 'M'
	jne .islastumb?		; last or corrupted -->
	inc ax
	add ax, word [ bx + 3 ]
	jmp short .loopumb	; process next -->
.islastumb?:
	cmp byte [ bx ], 'Z'
	jne .none		; corrupted -->
	jcxz .none		; didn't find that UMCB -->
			; The MCB at dx which was behind the one that contained the 'Z'
			; signature when we disabled the UMB link is now a valid MCB in
			; the MCB chain after we enabled the UMB link. All previous MCBs
			; are now 'M'.
	mov word [ cs:firstumcb ], dx
.none:
.got:
				; no need to restore es here since we do that below
				; no need to restore ds since we set it below

scancl:			; bp = selected options
	xor bp, bp		; initial option flags

	push cs
	pop es
	push cs
	pop ds			; ds, es changed while searching UMCB
	mov si, 81h		; PSP command line
.loop:	lodsb
	cmp al, 9		; tab ?
	je .loop
	cmp al, 32		; blank ?
	je .loop
	jb .end			; zero, cr, lf ? -->
	cmp al, '-'
	je .switch
	cmp al, '/'
	je .switch

	call getnybble
	jc .invalid
	xor bx, bx
	mov bl, al
	lodsb
	call getnybble
	jc .check_end_number
	add bx, bx
	add bx, bx
	add bx, bx
	add bx, bx
	add bl, al
	lodsb
.check_end_number:
	cmp al, 9
	je @F
	cmp al, 32
	ja .invalid
@@:

.number:
	dec si
	test bp, _got1stnumber
	jnz @F
	or bp, _got1stnumber
	mov byte [firstnumber], bl
	jmp .loop

@@:
	test bp, _got2ndnumber
	jnz @F
	or bp, _got2ndnumber
	mov byte [secondnumber], bl
	jmp .loop

@@:
	mov dx, msg.toomanynumbers
	jmp abort_msg


			; parse switch
.switch:
	lodsb			; get next character
	call capitalise
			; check against known switches
	push si
	mov si, parameters
	xchg al, ah
.loopswitch:
	lodsb
	xchg ax, cx
	lodsb
	xchg ax, cx		; cx = option flags
	cmp al, 0FFh
	je .invalid		; not found!
	cmp al, ah
	jne .loopswitch		; try next -->
	pop si
	xor ch, ch
	or bp, cx
	jmp .loop		; ds:si-> character after switch character. just continue


.invalid:			; display help if invalid
	mov dx, msg.help
	jmp abort_msg


.end:


findinstalleddebugger:
	mov ah, 0FFh		; start with multiplex number 0FFh
.loop:
	call .check
	mov al, 30h		; al = 30h to indicate found, ah = multiplex number
	jnc .end
	sub ah, 1		; search is backward (to find latest installed first), from 0FFh to 00h including
	jnc .loop		; try next if we didn't check all yet -->

	xor ax, ax		; al = 0 to indicate none found
	jmp .end		;  If not found, continue -->


		; INP:	ah = multiplex number to check
		;	ds = ss = cs
		; OUT:	CY if multiplex number unused or no signature match,
		;	 bp, ah, ds unmodified
		;	NC if match found,
		;	 ah = multiplex number (unmodified)
		; CHG:	si, di, es, cx, dx
.check:
	mov al, 00h		; AMIS installation check
	int 2Dh			; AMIS (or "DOS reserved" = iret if no AMIS present)
	cmp al, 0FFh
	jne .notfound
	mov si, debuggeramissig	; ds:si -> our AMIS name strings
	mov es, dx		; es:di -> name strings of AMIS multiplexer that just answered
	mov cx, 8		; Ignore description, only compare vendor and program name
	repe cmpsw
	je .checkret		; ZR, NC = match -->
.notfound:
	stc			; NZ, CY no match
.checkret:
	retn

.end:
	mov word [debuggerfunction], ax


process:
	test bp, _got1stnumber
	jnz @F
	mov dx, msg.nonumber
.abort_msg:
	jmp abort_msg

@@:
	mov dx, msg.1st
	call disp_msg

	mov ah, byte [firstnumber]
	mov al, 00h
	int 2Dh
	cmp al, 0
	je .free
	cmp al, 0FFh
	je .inuse
.unknown:
	mov dx, msg.unknown
	jmp @F

.free:
	mov dx, msg.free
@@:
	test bp, _got2ndnumber
	jnz .abort_msg
	call disp_msg
	jmp ..@exit

.inuse:
	mov ds, dx
	mov dx, msg.inuse.1
	call disp_msg
	call disp_amis_sig
	mov dx, msg.inuse.2
	call disp_msg
	call disp_amis_sig
	mov dx, msg.inuse.3
	call disp_msg

	mov al, 04h
	xor dx, dx
	xor bx, bx		; bl = 0 (interrupt to query), dx:bx = 0
	int 2Dh
	cmp al, 04h
	je @F
.nolist:
	mov dx, msg.nolist
	jmp .abort_msg

@@:
	mov ax, dx
	or dx, bx		; is dx:bx == 0 ?
	jz .nolist		; yes, invalid -->
	mov dx, msg.list.1
	call disp_msg
	call disp_ax_hex
	mov dx, msg.list.2
	call disp_msg
	mov es, ax
	mov ax, bx
	call disp_ax_hex
	mov dx, msg.list.3
	call disp_msg

countlist:
	mov di, bx
	xor cx, cx
	mov al, 2Dh
.loop:
	scasb
	je .found
	inc cx
		; At 18 bytes per IISP entry,
		;  4096 * 18 already overflows a 64 KiB segment.
	cmp cx, 1000h
	ja .toolong
	inc di
	inc di
	jmp .loop

.toolong:
	mov dx, msg.toolong
	jmp abort_msg

.found:
	inc cx
	mov ax, cx
	cmp ax, 1
	jne @F

	mov dx, msg.length.1.singular
	call disp_msg
	call disp_ax_dec
	mov dx, msg.length.2.singular
	call disp_msg
	jmp @FF

@@:
	mov dx, msg.length.1.plural
	call disp_msg
	call disp_ax_dec
	mov dx, msg.length.2.plural
	call disp_msg
@@:

determinetype:
	mov bx, word [es:di]
	mov dx, msg.entry.1
	call disp_msg
	mov ax, es
	call disp_ax_hex
	mov dx, msg.entry.2
	call disp_msg
	mov ax, bx
	call disp_ax_hex
	mov dx, msg.entry.3
	call disp_msg

	call IsIISPEntry?
	je @F
	mov dx, msg.entry.noiisp
	jmp abort_msg

@@:
	cmp word [es:bx + ieEntry], 010EBh
	jne .notstd
	cmp byte [es:bx + ieJmphwreset], 0EBh
	jne .notstd
.std:
	mov dx, msg.std
	jmp .gotstd

.notstd:
	cmp word [es:bx], 0EA90h
	mov dx, msg.uninstall
	je .gotstd
	mov dx, msg.nonstd.head
	call disp_msg
	cmp word [es:bx + ieEntry], 010EBh
	je @F
	mov dx, msg.nonstd.entry
	call disp_msg
@@:
	cmp byte [es:bx + ieJmphwreset], 0EBh
	je @F
	mov dx, msg.nonstd.hwreset
	call disp_msg
@@:
	mov dx, msg.crlf

.gotstd:
	call disp_msg

determinesettable:
	cmp byte [es:bx], 0EBh
	jne .notset
	inc bx
	mov al, [es:bx]
	inc bx
	cbw
	add bx, ax
	cmp bx, -16
	ja .notset
	mov di, bx
	push cs
	pop ds
	mov si, sequence
	mov cx, sequence.length
	mov al, byte [firstnumber]
	mov byte [si + sequence.imm8 - sequence], al
	repe cmpsb
	je .set
.notset:
	mov dx, msg.notset
	test bp, _got2ndnumber
	jnz abort_msg
	call disp_msg
	jmp ..@exit

.set:
	mov dx, msg.set.1
	call disp_msg
	mov ax, bx
	call disp_ax_hex
	mov dx, msg.set.2
	call disp_msg

setnumber:
	test bp, _got2ndnumber
	jz .exit
	mov ah, byte [secondnumber]
	mov al, 00h
	int 2Dh
	mov dx, msg.setting
	cmp al, 00h
	je .setit
	test bp, _force
	jz .notfree
	mov dx, msg.settingforce
.setit:
	call disp_msg
	mov byte [es:bx + sequence.imm8 - sequence], ah

.exit:
	jmp ..@exit

.notfree:
	mov dx, msg.notsetting
	jmp abort_msg


sequence:
.:
	cmp ah, 0FFh
.imm8: equ $ - 1
	je .
.length: equ $ - 1 - .


disp_amis_sig:
	push ax
	mov dx, di
	add di, 8
	push di
	db __TEST_IMM8		; (skip dec)
@@:
	dec di
	cmp dx, di
	je @F
	cmp byte [di - 1], 32
	je @B
@@:
	mov cx, di
	pop di
	sub cx, dx
	mov ah, 40h
	mov bx, 1
	int 21h
	pop ax
	retn


capitalise:
	cmp al, 'a'
	jb .notlowercase
	cmp al, 'z'
	ja .notlowercase
	and al, ~ 20h		; Uppercase if required
.notlowercase:
	retn


getnybble:
	cmp al, '0'
	jb .invalid		; unknown character -->
	cmp al, '9'
	jbe @F
	call capitalise
	cmp al, 'A'
	jb .invalid
	cmp al, 'F'
	ja .invalid
	sub al, 'A' - 10 - '0'
@@:
	sub al, '0'
	db __TEST_IMM8		; (NC, skip stc)
.invalid:
	stc
	retn


getdecit:
	cmp al, '0'
	jb .invalid		; unknown character -->
	cmp al, '9'
	ja .invalid
	sub al, '0'
	db __TEST_IMM8		; (NC, skip stc)
.invalid:
	stc
	retn


..@exit:				; terminate successful, re-used by some code (install too)
	call restorestate
	mov ax, 4C00h
	int 21h


abort_msg:
	call disp_msg
abort:
	call restorestate
	mov ax, 4CFFh
	int 21h


		; INP:	es:bx-> interrupt entry
		; OUT:	NZ if non-IISP entry
		;	ZR if IISP entry
IsIISPEntry?:
	cmp bx, - (ieSignature + 2)		; may access word at offset FFFFh ?
	ja .return				; yes, avoid --> (NZ)
	cmp word [ es:bx + ieSignature ], "KB"	; "KB"/424Bh ? ("BK" in MASM)
	jne .return
	cmp word [ es:bx + ieEntry ], 0EA90h	; nop\jmp far imm16:imm16 ?
	je .return				; unused IISP entry (created by iHPFS) -->
	cmp byte [ es:bx + ieEntry ], 0EBh	; jmp short ... ?
		; (This opcode should strictly be jmp short $+18 but there's programs
		; that save an additional jmp opcode by jumping directly into their
		; code even though it's not right behind the header.)
	jne .return
	cmp byte [ es:bx + ieJmphwreset ], 0EBh	; jmp short ... ?
	je .return				; usual IISP entry -->
	cmp byte [ es:bx + ieJmphwreset ], 0CBh	; retf ?
	je .return				; a shorter variant -->
	cmp byte [ es:bx + ieJmphwreset ], 0CFh	; iret ?
.return:
	retn


i24:
	mov al, 3			; always return fail, to handle the error as a soft one
	iret

i23:
	mov word [ cs:$ ], (__JMP_REL8|__REL16__(.return)<<8)	; don't reenter
	call restorestate
.return:
	stc				; always abort program (what default DOS handler also does)
retf_inst:
	retf

		; Restore modified DOS data
		;
		; CHG:	-
		; USE:	Int21
restorestate:
	push ax
	push bx
	mov bx, word [ cs:restored.strategy ]
	mov ax, 5801h
	int 21h				; restore strategy

	mov bx, word [ cs:restored.umblink ]
	mov ax, 5803h
	int 21h				; restore UMB link
	pop bx
	pop ax
	retn


disp_msg:
	 push ds
	 push ax
	push cs
	pop ds
	mov ah, 09h
	int 21h
	 pop ax
	 pop ds
	retn


disp_ax_hex:
	xchg al, ah
	call disp_al_hex
	xchg al, ah
disp_al_hex:
		push cx
		mov cl, 4
		ror al, cl
		call .nibble
		ror al, cl
		pop cx
.nibble:
		push ax
		and al, 0Fh
		add al, '0'
		cmp al, '9'
		jbe .isdigit
		add al, 'A'-('9'+1)
.isdigit:
		xchg dl, al
		mov ah, 02h
		int 21h
		pop ax
		retn


	; Following call: Display number in ax decimal
	; all registers preserved except dx
disp_ax_dec:			; ax (no leading zeros)
	; In: number in ax
	; Out: displayed
		push bx
		xor bx, bx
.pushax:
		push ax
.pushend:
		or bl, bl
		jz .nobl
		sub bl, 5
		neg bl
.nobl:
		push cx
		mov cx, 10000
		call .divide_out
		mov cx, 1000
		call .divide_out
		mov cx, 100
		call .divide_out
		mov cl, 10
		call .divide_out
							; (Divisor 1 is useless)
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h
		pop cx
		pop ax
		pop bx					; Caller's register
		retn

.divide_out:
	; In: ax = number
	;     cx = divisor
	; Out: ax = remainder of operation
	;      result displayed
		push dx
		xor dx, dx
		div cx				; 0:ax / cx
		push dx				; remainder
		dec bl
		jnz .nobl2
		or bh, 1
.nobl2:
		or bh, al
		jz .leadingzero
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h				; display result
 .leadingzero:
		pop ax				; remainder
		pop dx
		retn


	align 2, db 0
firstumcb:	dw 9FFFh			; guess of where UMBs start (A000h+)
firstmcb:	dw 0
restored:
.umblink:	dw 0
.strategy:	dw 0
debuggerfunction:
		dw 0				; = 0 if unused
firstnumber:	db 0
secondnumber:	db 0

	align 2, db 0
parameters:			; Valid switches on the command line, followed
				; by a byte with one option flag set.
	db "F", _force
	db 0FFh			; table end marker
			; The help switches are not in this table (and neither in the option
			; flags) because on invalid switches, we'll show the help anyway.


	align 2, db 0
debuggeramissig:
.ven:	fill 8,32,db "ecm"	; vendor
.prod:	fill 8,32,db "lDebug"	; product
msg:
.name:		ascic "RENUMBER: "
.help:		db "Change resident AMIS multiplexer's multiplex number",13,10
		db "Free software by C. Masloch",13,10
		db "Parameters:",13,10
		db "        NN       Operate on multiplex number NNh (source)",13,10
		db "        MM       Change to multiplex number MMh (destination)",13,10
		db "Options:",13,10
		db "        /f       Force even if MMh already in use",13,10
		db 13,10
		db "The switch character - is also accepted."
.crlf:		db 13,10,36
.toomanynumbers:ascic "Too many numbers specified!",13,10
.nonumber:	ascic "No number specified!",13,10
.1st:		ascic "1st number "
.unknown:	ascic "returned unknown status!",13,10
.free:		ascic "is free.",13,10
.inuse.1:	ascic "is used by vendor ",34
.inuse.2:	ascic 34," product ",34
.inuse.3:	ascic 34,".",13,10
.nolist:	ascic "Multiplexer did not return an interrupt list!",13,10
.list.1:	ascic "Multiplexer returned interrupt list at "
.entry.2:
.list.2:	ascic "h:"
.set.2:
.entry.3:
.list.3:	ascic "h.",13,10
.toolong:	ascic "Interrupt list is too long!",13,10
.length.1.singular:
.length.1.plural:	ascic "Interrupt list has "
.length.2.singular:	ascic " entry. "
.length.2.plural:	ascic " entries. "
.entry.1:	ascic "Interrupt 2Dh entry at "
.entry.noiisp:	ascic "Entry does not have an IISP header!",13,10
.std:		ascic "Standard IISP header.",13,10
.uninstall:	ascic "Uninstalled IISP header.",13,10
.nonstd.head:	ascic "Non-standard IISP header."
.nonstd.entry:	ascic " Entry is non-standard."
.nonstd.hwreset:ascic " HW reset is non-standard."
.set.1:		ascic "Can set number, cmp ah instruction at "
.notset:	ascic "Cannot set number!",13,10
.setting:	ascic "Setting to 2nd number.",13,10
.settingforce:	ascic "Setting to 2nd number even though it is not free!",13,10
.notsetting:	ascic "Not setting to 2nd number as it is not free!",13,10
